#include "Mesh.h"

#include "BspNode.h"

#include <iostream>

// We use the tiny_obj_loader library to load data of .obj-files into this mesh:
#define TINYOBJLOADER_IMPLEMENTATION

// We want to use triangulation when loading non-triangulated polygons:
#define TINYOBJLOADER_USE_MAPBOX_EARCUT

// Include tiny_obj_loader:
#include "tiny_obj_loader.h"

Mesh::Mesh()
{
}

Mesh::Mesh(std::vector<Triangle> triangles) : triangles(triangles) {
    computeBoundingBox();
    createBuffers();
}

/** 
 * @brief This contructor loads an obj file into the VAO and VBO (see header)
 */
Mesh::Mesh(std::string filepath){
    tinyobj::attrib_t attrib;
    std::vector<tinyobj::shape_t> shapes;
    std::vector<tinyobj::material_t> materials;
    std::string warn, err;
    bool retval = tinyobj::LoadObj( &attrib, &shapes, &materials, &warn, &err, filepath.c_str() );
    if (!warn.empty())
        std::cout << "[Mesh] Warning in LoadObj: " << warn << std::endl;
    if (!err.empty())
        std::cout << "[Mesh] Error in LoadObj: " << err << std::endl;
    if (!retval)
        std::cout << "[Mesh] LoadObj failed!" << std::endl;
    std::cout << "[Mesh] Loading " << filepath << std::endl;

    std::vector<glm::vec4> vertices_;
    std::vector<glm::vec4> normals;
    for (int i=0; i+2<attrib.vertices.size(); i+=3)
        vertices_.push_back( glm::vec4(attrib.vertices[i+0],
                                       attrib.vertices[i+1],
                                       attrib.vertices[i+2],
                                       1 ) );
    for (int i=0; i+2<attrib.normals.size(); i+=3)
        normals.push_back( glm::normalize(
                             glm::vec4(attrib.normals[i+0],
                                       attrib.normals[i+1],
                                       attrib.normals[i+2],
                                       0 )) );
    for (int i=0; i<shapes.size(); ++i) {
        for (int j=0; j+2<shapes[i].mesh.indices.size(); j+=3) {
            tinyobj::index_t i0 = shapes[i].mesh.indices[j+0];
            tinyobj::index_t i1 = shapes[i].mesh.indices[j+1];
            tinyobj::index_t i2 = shapes[i].mesh.indices[j+2];
            Triangle triangle;
            triangle.a.position = vertices_[i0.vertex_index];
            triangle.b.position = vertices_[i1.vertex_index];
            triangle.c.position = vertices_[i2.vertex_index];
            triangle.a.normal = normals[i0.normal_index];
            triangle.b.normal = normals[i1.normal_index];
            triangle.c.normal = normals[i2.normal_index];
            triangles.push_back( triangle );
        }
    }

    computeBoundingBox();
    createBuffers();
}

/**
 * Generates and binds buffers for vertex and index data, and sets up vertex
 * attribute pointers for a mesh.
 */
void Mesh::createBuffers() {
    glGenBuffers(1, &vbo);
    glGenBuffers(1, &vio);

    // std::vector<Vertex> vboData;
    unsigned int n = 0;
    for (; n < triangles.size();) {
        for (int i = 0; i < 2; ++i) {
            vertices.push_back(*(&triangles[n].a + i + (i % 2 == 0 ? 1 : -1)));
        }
        vertices.push_back(*(&triangles[n].c));
        ++n;
    }

    std::vector<unsigned int> indices;
    for (unsigned int i = 0; i < vertices.size(); ++i) {
        indices.push_back(i % 3 == 0 ? i + 1 : i % 3 == 1 ? i - 1 : i);
    }

    triangles.clear();
    for (int i=0; i<indices.size(); i+=3) {
        Triangle tri( vertices[indices[i+0]], vertices[indices[i+1]], vertices[indices[i+2]] );
        triangles.push_back( tri );
    }

    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vio);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned int), &indices[0], GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(Vertex), &vertices[0], GL_STATIC_DRAW);
    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), 0);
    glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(sizeof(glm::vec3)));

    for (int i = 0; i < 2; ++i)
        glEnableVertexAttribArray(i);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
}

void Mesh::clearBuffers()
{
    vertices.clear();
    if(vao != 0) glDeleteVertexArrays(1, &vao);
    if(vbo != 0) glDeleteBuffers(1, &vbo);
    if(vio != 0) glDeleteBuffers(1, &vio);
    vao = 0;
    vbo = 0;
    vio = 0;
}

void Mesh::clearBuffersBsp()
{
    verticesBsp.clear();
    if(vaoBsp != 0) glDeleteVertexArrays(1, &vaoBsp);
    if(vboBsp != 0) glDeleteBuffers(1, &vboBsp);
    if(vioBsp != 0) glDeleteBuffers(1, &vioBsp);
    vaoBsp = 0;
    vboBsp = 0;
    vioBsp = 0;
}

void Mesh::computeBoundingBox()
{
    for (Triangle t : triangles) {
        for (Vertex v : {t.a, t.b, t.c}) {
            m_bbLo.x = std::min(m_bbLo.x, v.position.x);
            m_bbLo.y = std::min(m_bbLo.y, v.position.y);
            m_bbLo.z = std::min(m_bbLo.z, v.position.z);
            m_bbHi.x = std::max(m_bbHi.x, v.position.x);
            m_bbHi.y = std::max(m_bbHi.y, v.position.y);
            m_bbHi.z = std::max(m_bbHi.z, v.position.z);
        }   
    }
}

Mesh::~Mesh() {
    clearBuffers();
    clearBuffersBsp();
}

void Mesh::render( Shader &shader )
{
    // The fragment shader determines the fragment color based on
    // lighting and object color.
    shader.setUniform( "bbMin", m_bbLo );
    shader.setUniform( "bbMax", m_bbHi );
    shader.setUniform( "objectColor", m_color );
    shader.setUniform( "shininess", m_shininess );
    shader.setUniform( "specularStrength", m_specularStrength );
    shader.setUniform( "alpha", 0.5f );

    shader.bind();

    // Draws the geometry data that is defined by the vao:
    glBindVertexArray(vao);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vio);
    glDrawElements(GL_TRIANGLES, (GLsizei)(triangles.size()*3), GL_UNSIGNED_INT, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
}

void Mesh::render( Shader &shader, glm::vec3 eyePosition )
{
    // The fragment shader determines the fragment color based on
    // lighting and object color.
    shader.setUniform( "bbMin", m_bbLo );
    shader.setUniform( "bbMax", m_bbHi );
    shader.setUniform( "objectColor", m_color );
    shader.setUniform( "shininess", m_shininess );
    shader.setUniform( "specularStrength", m_specularStrength );
    shader.setUniform( "alpha", 0.5f );

    shader.bind();

    if (m_renderMode == RenderMode::Unsorted) {
        // *** Default rendering ***
        render( shader );
        m_triangleCount = triangles.size();
        return;
    }
    
    // *** Sort the triangle rendering order using BSP ***
    std::vector<Triangle> tris;
    if (m_renderMode == RenderMode::BspNoSplit) {
        bspRootNodeNoSplit->gather( tris, eyePosition );
    } else {
        bspRootNode->gather( tris, eyePosition );
    }
    m_triangleCount = tris.size();

    // *** Render the sorted triangles ***
    // Prepare buffers
    clearBuffersBsp();
    glGenBuffers(1, &vboBsp);
    glGenBuffers(1, &vioBsp);
    for (int n=0; n < tris.size(); ++n) {
        for (int i = 0; i < 2; ++i) {
            verticesBsp.push_back(*(&tris[n].a + i + (i % 2 == 0 ? 1 : -1)));
        }
        verticesBsp.push_back(*(&tris[n].c));
    }

    std::vector<unsigned int> indicesBsp;
    for (unsigned int i = 0; i < verticesBsp.size(); ++i) {
        indicesBsp.push_back(i % 3 == 0 ? i + 1 : i % 3 == 1 ? i - 1 : i);
    }

    // Transfer data to GPU
    glGenVertexArrays(1, &vaoBsp);
    glBindVertexArray(vaoBsp);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vioBsp);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indicesBsp.size() * sizeof(unsigned int), &indicesBsp[0], GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, vboBsp);
    glBufferData(GL_ARRAY_BUFFER, verticesBsp.size() * sizeof(Vertex), &verticesBsp[0], GL_STATIC_DRAW);
    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), 0);
    glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(sizeof(glm::vec3)));

    for (int i = 0; i < 2; ++i)
        glEnableVertexAttribArray(i);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindVertexArray(0);

    // Draw call for the newly created buffer
    glBindVertexArray(vaoBsp);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vioBsp);
    glDrawElements(GL_TRIANGLES, (GLsizei)(tris.size()*3), GL_UNSIGNED_INT, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
}

void Mesh::updateBsp()
{
    delete bspRootNode;
    delete bspRootNodeNoSplit;
    bspRootNode = new BspNode( triangles, true );
    bspRootNodeNoSplit = new BspNode( triangles, false );
}