// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include this file only once when compiling:
#pragma once

// Include the std::vector class, which is comparable to the
// ArrayList in Java (this is not a mathematical vector!):
#include <vector>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include matrix and vector:
#include <glm/glm.hpp>

#include <string>

// Include our Triangle primitive:
#include "Triangle.h"

#include "Renderable.h"
#include "Shader.h"
#include "BspNode.h"

enum RenderMode {
    Unsorted = 0,
    BspNoSplit = 1,
    BspSplit = 2,
};

/**
 * Represents a renderable mesh.
 */
class Mesh : public Renderable
{
public:
    Mesh();

    /**
     * Constructs a mesh which consists of the given vertices
     * and prepare for rendering (e.g. upload the data to the GPU).
     */
    Mesh(std::vector<Triangle> triangles);

    /**
     * Loads the geometry data of an .obj file into this mesh object.
     */
    Mesh(std::string filepath);

    /**
     * Deletes the mesh and ensures that the memory on the GPU
     * is cleaned up.
     */
    ~Mesh();

    /**
     * Renders the mesh.
     */
    virtual void render( Shader &shader ) override;

    /**
     * Renders the mesh, sorted for correct transparency at eye position.
     */
    virtual void render( Shader &shader, glm::vec3 eyePosition ) override;

    /**
     * Creates buffers (VAO,VBO,etc.) at the GPU and uploads triangle
     * data. In previous frameworks, this was done directly in the
     * constructor, but it was splitted to be use with different
     * constructors.
     */
    void createBuffers();
    void clearBuffers();

    void createBuffersBsp();
    void clearBuffersBsp();


    GLuint vao = 0;
    GLuint vbo = 0;
    GLuint vio = 0;

    std::vector<Triangle> triangles;
    std::vector<Vertex> vertices;

    GLuint vaoBsp = 0;
    GLuint vboBsp = 0;
    GLuint vioBsp = 0;

    std::vector<Triangle> trianglesBsp;
    std::vector<Vertex> verticesBsp;

    /** The body's color in RGB in the range [0,1]^4. */
    glm::vec3 m_color = glm::vec3( 1.0, 0.0, 1.0 );

    float m_shininess = 1.0f;

    float m_specularStrength = 0.333f;

    glm::vec3 m_bbLo = glm::vec3((float)+FLT_MAX);
    glm::vec3 m_bbHi = glm::vec3((float)-FLT_MAX);

    int m_triangleCount = 0;
    RenderMode m_renderMode = RenderMode::Unsorted;
    BspNode *bspRootNode = nullptr;
    BspNode *bspRootNodeNoSplit = nullptr;
    void updateBsp();

private:
    void computeBoundingBox();
};
